@@ -1988,10 +1988,18 @@ reg_scan_mark_refs (x, insn, note_flag, min_regno)
      int note_flag;
      int min_regno;
 {
-  register enum rtx_code code = GET_CODE (x);
+  register enum rtx_code code;
   register rtx dest;
   register rtx note;
 
+  /* This can happen when scanning insns referenced by certain notes.
+
+     It is unclear if we should be scanning such insns; until someone can
+     say for sure this seems like the safest fix.  */
+  if (x == NULL_RTX)
+    return;
+
+  code = GET_CODE (x);
   switch (code)
     {
     case CONST_INT:
